/*
 * sverb: an order 15 CFDN reverb
 * http://sed.free.fr/sverb
 * public domain
 * (there may be some patent for this stuff, I don't know and I don't care)
 * Cdric Roux <sed@free.fr>
 *
 * Algorithm comes from:
 * "Maximally Diffusive Yet Efficient Feedback Delay Networks for
 *  Artificial Reverberation", Davide Rocchesso, IEEE SIGNAL
 * PROCESSING LETTERS, VOL. 4, NO. 9, SEPTEMBER 1997
 * http://www.scienze.univr.it/documenti//ArticoloRivista/allegato/
 *   allegato128605.pdf
 * And:
 * "Circulant and elliptic feedback delay networks for artificial
 *  reverberation", D. Rocchesso and J. O. Smith, IEEE Trans. Speech
 * Audio Processing, vol. 5, pp. 51 63, Jan. 1997.
 * http://ccrma.stanford.edu/~jos/cfdn/cfdn.pdf
 * And:
 * http://ccrma.stanford.edu/~jos/Reverb/First_Order_Delay_Filter_Design.html
 * http://ccrma.stanford.edu/~jos/Reverb/Tonal_Correction_Filter.html
 * for the loss filters
 */

/*
 * for the loss delay lines:
 * implements the function:
 * H(z) = g * (1 - a) / (1 - a * z-1)
 * Or (see http://ccrma.stanford.edu/~jos/filters/
              Z_Transform_Difference_Equations.html):
 * y(n) = g * (1 - a) * x(n) + a * y(n-1)
 *
 * g = pow(10, -3 * M * T / t60(0))
 *      with M = delay length in sample
 *       and T = sample period (1/44100 for 44.1 kHz audio signal)
 * a = ln(10) / 4 * log10(g) * (1 - 1/(alpha * alpha)
 *     with alpha = t60(pi/T) / t60(0)
 * t60(0) is desired decay time for freq 0
 * t60(pi/T) is desired decay time for quantizationfreq of audio signal
 *           (I guess) (so here it would be 44100 Hz)
 * c1 = g * (1 - a)
 * c2 = a
 */

/*
 * for the tonal coorection:
 * implements the function:
 * E(z) = (1 - b * z-1) / (1 - b)
 * Or (see http://ccrma.stanford.edu/~jos/filters/
              Z_Transform_Difference_Equations.html):
 * y(n) = 1/(1-b) * (x(n) - b * x(n-1));
 *
 * b = (1 - alpha) / (1 + alpha)
 *     with alpha = t60(pi/T) / t60(0)
 * t60(0) is desired decay time for freq 0
 * t60(pi/T) is desired decay time for quantizationfreq of audio signal
 *           (I guess) (so here it would be 44100 Hz)
 * c1 = b
 * c2 = 1/(1-b)
 */

#include "sverb.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "sverb_version.h"

extern long _sverb_asm_process(sverb *, long);
extern long _sverb_asm_process_no_t60(sverb *, long);

static void tonal_correction_init(tonal_correction_filter *f,
    float t60_0, float t60_pi)
{
  float alpha = t60_pi / t60_0;
  float b = (1. - alpha) / (1. + alpha);

  f->c1 = b;
  f->c2 = 1. / (1. - b);
  f->ox = 0;
}

static void tonal_correction_int_init(tonal_correction_filter_int *f,
    float t60_0, float t60_pi, int nb_bits)
{
  float alpha = t60_pi / t60_0;
  float b = (1. - alpha) / (1. + alpha);

  f->c2 = (long)((1. / (1. - b)) * (float)(1<<nb_bits));
  f->c1 = (long)((b / (1. - b)) * (float)(1<<nb_bits));
  f->ox = 0;
}

static float tonal_correction_process(tonal_correction_filter *f, float x)
{
  float y = f->c2 * (x - f->c1 * f->ox);
  f->ox = x;
  return y;
}

static long tonal_correction_process_int(tonal_correction_filter_int *f,
    long x, int nb_bits)
{
  long y = f->c2 * x - f->c1 * f->ox;
  f->ox = x;
  return y>>nb_bits;
}

/* T is in fact 1/T, for example you give 44100, not 1/44100 */
static void loss_delay_line_init(loss_delay_line_filter *f, int M, int T,
    float t60_0, float t60_pi)
{
  float alpha = t60_pi / t60_0;
  float g = pow(10., -3. * M / (t60_0 * T));
  float a = log(10) / 4. * log10(g) * (1 - 1 / (alpha * alpha));

  f->c1 = g * (1 - a);
  f->c2 = a;
  f->oy = 0;
}

static float loss_delay_line_process(loss_delay_line_filter *f, float x)
{
  float y = f->c1 * x + f->c2 * f->oy;
  f->oy = y;
  return y;
}

/* T is in fact 1/T, for example you give 44100, not 1/44100 */
static void loss_delay_line_int_init(loss_delay_line_filter_int *f, int M,
    int T, float t60_0, float t60_pi, int nb_bits)
{
  float alpha = t60_pi / t60_0;
  float g = pow(10., -3. * M / (t60_0 * T));
  float a = log(10) / 4. * log10(g) * (1 - 1 / (alpha * alpha));

  f->c1 = (long)((g * (1 - a)) * (float)(1<<nb_bits));
  f->c2 = (long)(a * (float)(1<<nb_bits));
  f->oy = 0;
}

static long loss_delay_line_process_int(loss_delay_line_filter_int *f,
    long x, int nb_bits)
{
  long y = f->c1 * x +
           f->c2 * f->oy;
  f->oy = y>>nb_bits;
  return f->oy;
}

/* this is the diffusion matrix (this implementation is too slow) */
static int turn(int i, int j)
{
  static int matrix[]= {
    0, 1, 1, 1, 0, 1, 1, 0, 0, 1, 0, 1, 0, 0, 0
  };
  j += i; if (j >= 15) j-=15;
  return matrix[j];
}

static float float_process(sverb *s, float x)
{
  int i, j;
  float y;
  float v[15];
  float ov[15];

  /* get y */
  y = 0;
  for (i=0; i<15; i++)
    y += s->delays[i][s->cur[i]];
  y = tonal_correction_process(&s->tc, y);

  /* read the tap */
  for (i=0; i<15; i++)
    ov[i] = loss_delay_line_process(&s->ld[i], s->delays[i][s->cur[i]]);

  /* get new value */
  for (i=0; i<15; i++) {
    register float too = 0;
    for (j=0; j<15; j++) {
      if (turn(i, j))
        too += ov[j];
      else
        too -= ov[j];
    }
    v[i] = too/4 + x;
  }

  /* store new value */
  for (i=0; i<15; i++) {
    s->delays[i][s->cur[i]++] = v[i];
    if (s->cur[i] == s->DELAY[i]) s->cur[i] = 0;
  }

  return y;
}

static float float_process_no_t60(sverb *s, float x)
{
  int i, j;
  float y;
  float v[15];
  float ov[15];

  /* get y */
  y = 0;
  for (i=0; i<15; i++)
    y += s->delays[i][s->cur[i]];

  /* read the tap */
  for (i=0; i<15; i++)
    ov[i] = s->delays[i][s->cur[i]];

  /* get new value */
  for (i=0; i<15; i++) {
    register float too = 0;
    for (j=0; j<15; j++) {
      if (turn(i, j))
        too += ov[j];
      else
        too -= ov[j];
    }
    v[i] = too/4 + x;
  }

  /* store new value */
  for (i=0; i<15; i++) {
    s->delays[i][s->cur[i]++] = v[i];
    if (s->cur[i] == s->DELAY[i]) s->cur[i] = 0;
  }

  return y;
}

static float int_process(sverb *s, float xi)
{
  int i, j;
  long x = xi * (1<<s->integer_precision);
  long y;
  long v[15];
  long ov[15];

  /* get y */
  y = 0;
  for (i=0; i<15; i++)
    y += s->delays_int[i][s->cur[i]];
  y = tonal_correction_process_int(&s->tc_int, y, s->integer_precision);

  /* read the tap */
  for (i=0; i<15; i++)
    ov[i] = loss_delay_line_process_int(&s->ld_int[i],
        s->delays_int[i][s->cur[i]], s->integer_precision);

  /* get new value */
  for (i=0; i<15; i++) {
    register long too = 0;
    for (j=0; j<15; j++) {
      if (turn(i, j))
        too += ov[j];
      else
        too -= ov[j];
    }
    v[i] = (too>>2) + x;
  }

  /* store new value */
  for (i=0; i<15; i++) {
    s->delays_int[i][s->cur[i]++] = v[i];
    if (s->cur[i] == s->DELAY[i]) s->cur[i] = 0;
  }

  return (float)y/(1<<s->integer_precision);
}

static float int_process_no_t60(sverb *s, float xi)
{
  int i, j;
  long x = xi * (1<<s->integer_precision);
  long y;
  long v[15];
  long ov[15];

  /* get y */
  y = 0;
  for (i=0; i<15; i++)
    y += s->delays_int[i][s->cur[i]];

  /* read the tap */
  for (i=0; i<15; i++)
    ov[i] = s->delays_int[i][s->cur[i]];

  /* get new value */
  for (i=0; i<15; i++) {
    register long too = 0;
    for (j=0; j<15; j++) {
      if (turn(i, j))
        too += ov[j];
      else
        too -= ov[j];
    }
    v[i] = (too>>2) + x;
  }

  /* store new value */
  for (i=0; i<15; i++) {
    s->delays_int[i][s->cur[i]++] = v[i];
    if (s->cur[i] == s->DELAY[i]) s->cur[i] = 0;
  }

  return (float)y/(1<<s->integer_precision);
}

#ifdef ASM_IA32
static float asm_process(sverb *s, float xi)
{
  long x = xi * (1<<s->integer_precision);
  long y = _sverb_asm_process(s, x);
  return (float)y/(1<<s->integer_precision);
}

static float asm_process_no_t60(sverb *s, float xi)
{
  long x = xi * (1<<s->integer_precision);
  long y = _sverb_asm_process_no_t60(s, x);
  return (float)y/(1<<s->integer_precision);
}
#endif

static void set_function(sverb *s)
{
  if (s->type == SVERB_FLOAT) {
    if (s->bypass_t60)
      s->sverb_process = float_process_no_t60;
    else
      s->sverb_process = float_process;
  } else if (s->type == SVERB_INT) {
    if (s->bypass_t60)
      s->sverb_process = int_process_no_t60;
    else
      s->sverb_process = int_process;
#ifdef ASM_IA32
  } else {
    if (s->bypass_t60)
      s->sverb_process = asm_process_no_t60;
    else
      s->sverb_process = asm_process;
#endif
  }
}

void sverb_init(sverb *s)
{
  /* for use with the gui, we require that sizeof(float) = sizeof(int)
   * if you only use float or only int or only asm in your code,
   * you may remove this test
   */
  if (sizeof(float) != sizeof(int)) {
    fprintf(stderr,
        "error, sizeof(float) != sizeof(int), contact the authors (%s)\n",
        SVERB_EMAIL);
    exit(1);
  }
  /* I am lazy, so I want this (i should clean the code, but hey...). */
  if (sizeof(long) != sizeof(int)) {
    fprintf(stderr,
        "error, sizeof(long) != sizeof(int), contact the authors (%s)\n",
        SVERB_EMAIL);
    exit(1);
  }

  memset(s, 0, sizeof(sverb));
  s->sverb_process = float_process;
  s->type = SVERB_FLOAT;
  s->integer_precision = 10;
}

void sverb_reset(sverb *s)
{
  int i;

  s->tc_int.ox = 0;
  s->tc.ox = 0;

  for (i=0; i<15; i++) {
    s->cur[i] = 0;
    if (s->type == SVERB_FLOAT)
      memset(s->delays[i], 0, sizeof(float) * s->DELAY[i]);
    else
      memset(s->delays_int[i], 0, sizeof(int) * s->DELAY[i]);
    s->ld[i].oy = 0;
    s->ld_int[i].oy = 0;
  }
}

void sverb_free(sverb *s)
{
  int i;

  for (i=0; i<15; i++)
    if (s->type == SVERB_FLOAT)
      free(s->delays[i]);
    else
      free(s->delays_int[i]);

  memset(s, 0, sizeof(sverb));
}

void sverb_set_t60(sverb *s, float t60_0, float t60_pi)
{
  int i;

  s->t60_0 = t60_0;
  s->t60_pi = t60_pi;

  if (s->type == SVERB_FLOAT) {
    tonal_correction_init(&s->tc, t60_0, t60_pi);
    for (i=0; i<15; i++)
      loss_delay_line_init(&s->ld[i], s->DELAY[i], 44100, t60_0, t60_pi);
  } else {
    tonal_correction_int_init(&s->tc_int,
        t60_0, t60_pi, s->integer_precision);
    for (i=0; i<15; i++)
      loss_delay_line_int_init(&s->ld_int[i], s->DELAY[i], 44100,
          t60_0, t60_pi, s->integer_precision);
  }

  sverb_reset(s);
}

void sverb_set_type(sverb *s, int t)
{
  int i;

  /* reuse memory, that's why we need sizeof(float) = sizeof(int) */
  if (s->type == SVERB_FLOAT && t != SVERB_FLOAT)
    for (i=0; i<15; i++)
      s->delays_int[i] = (long *)s->delays[i];
  else if (s->type != SVERB_FLOAT && t == SVERB_FLOAT)
    for (i=0; i<15; i++)
      s->delays[i] = (float *)s->delays_int[i];

  s->type = t;
  sverb_set_t60(s, s->t60_0, s->t60_pi);
  sverb_reset(s);
  set_function(s);
}

void sverb_set_integer_precision(sverb *s, int p)
{
  s->integer_precision = p;
  if (s->type != SVERB_FLOAT)
    sverb_set_t60(s, s->t60_0, s->t60_pi);
  sverb_reset(s);
}

void sverb_use_t60(sverb *s, int use_t60)
{
  s->bypass_t60 = !use_t60;
  sverb_reset(s);
  set_function(s);
}

/* return -1 if error, 0 if ok */
int sverb_set_delay(sverb *s, int pos, int val)
{
  if (s->type == SVERB_FLOAT) {
    float *new = malloc(sizeof(float) * val);
    if (!new) return -1;
    s->delays[pos] = new;
  } else {
    long *new = malloc(sizeof(long) * val);
    if (!new) return -1;
    s->delays_int[pos] = new;
  }
  s->DELAY[pos] = val;
  s->cur[pos] = 0;
  return 0;
}
